'''
Created on 2021-12-08

@author: wf
'''
from spreadsheet.spreadsheet import SpreadSheet, SpreadSheetType


class TableEditing(object):
    '''
    table Editing
    
    enhancement
    onlinespreadsheet editing call
    validation
    '''
    def __init__(self,lods:dict=None):
        '''
        Constructor
        
        Args:
            lods(dict): a dict of list of dicts that represents the content of a Spreadsheet
        '''
        if lods is None:
            self.lods={}
        else:
            self.lods=lods
        self.enhanceCallbacks=[] # functions to call for enhancing
         
    def toSpreadSheet(self,spreadSheetType:SpreadSheetType,name:str)->SpreadSheet:
        '''
        convert me to the given spreadSheetType
        
        Args:
            spreadSheetType(SpreadSheetType): the type of onlinespreadsheet to create
            name(str): the name of the onlinespreadsheet
        '''
        spreadSheet=SpreadSheet.create(spreadSheetType,name=name)
        spreadSheet.tables=self.lods
        return spreadSheet
    
    def fromSpreadSheet(self, spreadSheet:SpreadSheet):
        pass
    
    def addLoD(self,name,lod):
        '''
        add the given list of dicts with the given name to my lods
        
        Args:
            name(str): the name 
            lod(list): the list of dicts to add
        '''
        self.lods[name]=lod
    
    def addEnhancer(self,callback):
        '''
        add the given enhancer callback to my callbacks
        
        Args:
            callback(func): the callback function to add
        '''
        self.enhanceCallbacks.append(callback)
    
    def enhance(self):
        '''
        enhance/enrich my list of dicts with the set callbacks
        '''
        for callback in self.enhanceCallbacks:
            callback(self)    


'''
Created on 2022-04-21

@author: wf
'''

class Version(object):
    '''
    Version handling for pyGenericSpreadSheet
    '''
    version='0.0.23'
    date = '2022-04-21'
    updated = '2022-07-27'
    name='pyGenericSpreadSheet'
    description='python API providing generic Access to Spreadsheet backends'

        



'''
Created on 2022-04-18

@author: wf
'''
import requests
from io import StringIO
import pandas as pd

class GoogleSheet(object):
    '''
    GoogleSheet Handling
    '''

    def __init__(self, url):
        '''
        Constructor
        '''
        self.url=url
        self.dfs={}
        
    def open(self,sheetNames):
        '''
        Args:
            sheets(list): a list of sheetnames
        '''
        self.sheetNames=sheetNames
        for sheetName in sheetNames:
            #csvurl=f"{self.url}/export?format=csv"
            csvurl=f"{self.url}/gviz/tq?tqx=out:csv&sheet={sheetName}"
            response=requests.get(csvurl)
            csvStr=response.content.decode('utf-8')
            self.dfs[sheetName]=pd.read_csv(StringIO(csvStr),keep_default_na=False)
        
    def asListOfDicts(self,sheetName):
        '''
        convert the given sheetName to a list of dicts
        
        Args:
            sheetName(str): the sheet to convert
        '''
        lod=self.dfs[sheetName].to_dict('records') 
        return lod
        
        
        
        
        

        


import io
import math
import os
import dateutil.parser as dateparser
import pandas as pd


from io import BytesIO
from enum import Enum,auto
from zipfile import ZipFile
from lodstorage.csv import CSV
from pandas import Timestamp, NaT
from datetime import datetime, date
from werkzeug.datastructures import FileStorage

    
class Format:
    '''
    potential Formats 
    '''
    formatMap={
        "CSV": {
            "name": "CSV",
            "title": "Comma separated Values",
            "postfix": ".csv",
            "mimetype": "text/csv"
            
        },
        "EXCEL": {
            "name": "Excel",
            "title": "Microsoft Excel",
            "postfix": ".xlsx",
            "mimetype": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
        },
        "JSON": {
            "name": "JSON",
            "title": "Javascript Simple Object Notation",
            "postfix": ".json",
            "mimetype": "application/json"
        },
        "ODS": {
            "name": "ODS",
            "title": "OpenDocument Spreadsheet",
            "postfix": ".ods",
            "mimetype": "application/vnd.oasis.opendocument.onlinespreadsheet"
        }
        
    }    
class SpreadSheetType(Enum):
    '''
    Entities of openresearch.
    Used to specify for a fixer the domain of operation.
    '''
    CSV=auto()
    EXCEL=auto()
    ODS=auto()
    JSON=auto()
    
    def getProperty(self,propertyName):
        value=Format.formatMap[self.name][propertyName]
        return value
       
    def getPostfix(self):
        return self.getProperty("postfix")
       
    def getName(self):
        return self.getProperty("name")
    
    def getMimeType(self):
        return self.getProperty("mimetype")
    
    def getTitle(self):
        return self.getProperty("title")
    
    @classmethod
    def asSelectFieldChoices(cls):
        choices=[]
        for i,choice in enumerate(cls):
            choices.append((choice.name,choice.getTitle()))
        return choices


class SpreadSheet:
    '''
    i am onlinespreadsheet
    '''

    FILE_TYPE=NotImplemented
    MIME_TYPE=NotImplemented
    
    def __init__(self,name:str,spreadSheetType:SpreadSheetType):
        '''
        constructor
        '''
        self.name=name
        self.spreadSheetType=spreadSheetType
        self.tables={} # dict of lods
        pass
    
    @classmethod
    def create(cls,spreadSheetType:SpreadSheetType,name:str):
        '''
        create a SpreadSheet of the given types
        
        Args:
            spreadSheetType(SpreadSheetType): the type of onlinespreadsheet to create
            name(str): the name of the onlinespreadsheet
        
        '''
        spreadSheet=None
        if spreadSheetType==SpreadSheetType.EXCEL:
            spreadSheet=ExcelDocument(name=name)
        elif spreadSheetType==SpreadSheetType.ODS:
            spreadSheet=OdsDocument(name=name)
        elif spreadSheetType==SpreadSheetType.CSV:
            spreadSheet=CSVSpreadSheet(name=name)
        return spreadSheet

    @classmethod
    def load(cls, document):
        """
        Tries to load the given document as SpreadSheet
        Args:
            document: onlinespreadsheet document to load

        Returns:
            SpreadSheet
        """
        documentSpreadSheetType=None
        documentName=""
        spreadsheet = None
        # TODO - use SpreadSheeeType enum instead
        spreadSheetTypes=[OdsDocument, ExcelDocument, CSVSpreadSheet]
        if isinstance(document, FileStorage):
            document.stream.seek(0)
            for spreadSheetType in spreadSheetTypes:
                if document.filename.endswith(spreadSheetType.FILE_TYPE):
                    documentName=document.filename[:-len(spreadSheetType.FILE_TYPE)]
                    documentSpreadSheetType=spreadSheetType
                    break
        elif isinstance(document, io.BytesIO) or isinstance(document, io.StringIO) or isinstance(document, io.TextIOWrapper):
            document.seek(0)
            for spreadSheetType in spreadSheetTypes:
                if document.name.endswith(spreadSheetType.FILE_TYPE):
                    documentName=document.name[:-len(spreadSheetType.FILE_TYPE)]
                    documentSpreadSheetType=spreadSheetType
                    break
        elif isinstance(document, str):
            try:
                buffer=None
                with open(document, "rb") as f:
                    buffer = io.BytesIO(f.read())
                    buffer.name=document
                return cls.load(buffer)
            except Exception as e:
                print(e)
                raise e
        else:
            print("Unable to load SpreadSheet")
            return None
        if documentSpreadSheetType:
            spreadsheet = documentSpreadSheetType(name=documentName)
            spreadsheet.loadFromFile(document)
        return spreadsheet
    
    def getTable(self, name:str):
        """
        returns the data corresponding to the given table name
        Args:
            name: name of the table

        Returns:
            LoD
        """
        if name in self.tables:
            return self.tables[name]
        
    def addTable(self, name:str, lod:list, headers:dict=None):
        """
        add the given data as table to the document

        Args:
            name(str): name of the table
            lod: data that should be added to the document as table
            headers(dict): Mapping from dict key to the new headers. Also functions as restriction. If not defined dict key are used as headers
        """
        if headers:
            lod=[{newHeader:record.get(oldHeader, None) for oldHeader, newHeader in headers.items()} for record in lod]
        self.tables[name]=lod

    def hasTable(self, name:str):
        """
        Checks if table under given name exists

        Args:
            name(str): name of the Table

        Retruns:
            True if table exists otherwise False
        """
        return name in self.tables

    def saveToFile(self, fileName:str=None, dir_name:str=None):
        """
        saves SpreadSheet to file

        Args:
            fileName(str): name of the file if None SpreadSheet name is used
            dir_name(str): name of directory to store the file

        Returns:
            Nothing
        """
        if fileName is None:
            fileName=self.filename
        if dir_name is not None:
            fileName=os.path.join(dir_name, fileName)
        documentBuffer=self.toBytesIO()
        with open(fileName, "wb") as f:
            documentBuffer.seek(0)
            f.write(documentBuffer.read())

    def toBytesIO(self) -> BytesIO:
        """
        Converts the document into an BytesIO stream

        Returns:
            BytesIO Stream of the document
        """
        raise NotImplementedError

    def loadFromFile(self, file, samples:dict=None):
        """
        Load SpreadSheet from given file or file object
        """
        tables=self._loadFromFile(file)
        if tables:
            for name,table in tables.items():
                if samples:
                    if name in samples:
                        self.fixLodTypes(table, samples[name])
                self.tables[name]=table
            
    def _loadFromFile(self, file):
        """
        load the document from the given .ods file
        Args:
            file: absolut file path to the file that should be loaded
            samples(dict): samples of the sheets. Expected format: sheetName:SamplesForSheet
        Returns:

        """
        if isinstance(file, str):
            try:
                with open(file, mode="rb") as f:
                    buffer=BytesIO()
                    buffer.write(f.read())
                    # work around along the line of
                    # https://stackoverflow.com/a/42811024/1497139
                    buffer.name=f.name
            except Exception as e:
                print(f"Tried to open {file} as a File and failed")
                raise e
        else:
            buffer=file
        return self._loadFromBuffer(buffer)

    def _loadFromBuffer(self, buffer):
        """
        Load SpreadSheet from given buffer

        Args:
            buffer: file like object

        """
        raise NotImplementedError

    @property
    def filename(self):
        return self.name + self.FILE_TYPE

    @staticmethod
    def fixLodTypes(lod: list, samples: list, typeConversionMap: dict = None):
        """
        Fixes the types of the values of the given lod by converting it to the type corresponding to the given sampeles

        Args:
            lod(list): List of dicts to be type fixed
            samples(list): list of samples specifying the value types
            typeConversionMap(dict): Map from type to corresponding conversion function. If None default conversions for string values are used.
        """
        if typeConversionMap is None:
            def toDate(value):
                if isinstance(value, str):
                    return dateparser.parse(value).date()
                elif isinstance(value, datetime):
                    return value.date()
                elif isinstance(value, date):
                    return value
                else:
                    print(f"{value} could not be converted to date")
                    return value
            typeConversionMap = {
                str:      lambda value: str(value),
                int:      lambda value: int(value),
                float:    lambda value: float(value),
                date:     lambda value: toDate(value),
                datetime: lambda value: dateparser.parse(value)
            }
        # build sample types map
        sampleTypes = {}
        for sample in samples:
            if isinstance(sample, dict):
                for key, value in sample.items():
                    valueType = type(value)
                    if key not in sampleTypes:
                        sampleTypes[key] = valueType
                    elif sampleTypes[key] != valueType:
                        print(f"Sample has inconsistent types for {key} the types {sampleTypes[key]} and {valueType} are defined")
                    else:
                        pass
        # fix types of lod
        for d in lod:
            if isinstance(d, dict):
                for key, value in d.items():
                    if value is not None and key in sampleTypes and sampleTypes[key] in typeConversionMap:
                        if type(value) != sampleTypes[key]:
                            d[key] = typeConversionMap[sampleTypes[key]](value)
            else:
                print("List of dicts contains a non dict item")

    
class CSVSpreadSheet(SpreadSheet):
    '''
    CSV Spreadsheet packaging as ZIP file of CSV files
    '''

    FILE_TYPE = '.zip'
    TABLE_TYPE = '.csv'
    MIME_TYPE = 'application/zip'

    def __init__(self, name: str):
        super().__init__(name=name, spreadSheetType=SpreadSheetType.CSV)

    def toBytesIO(self) -> BytesIO:
        """
        Converts the document into an BytesIO stream

        Returns:
            BytesIO Stream of the document
        """
        buffer = BytesIO()
        buffer.name=self.filename
        with ZipFile(buffer, mode="w") as documentZip:
            for tableName, table in self.tables.items():
                csv=CSV.toCSV(table)
                documentZip.writestr(tableName+self.TABLE_TYPE, csv)
        buffer.seek(0)
        return buffer

    def _loadFromBuffer(self, file):
        """
        load the document from the given .zip file
        Args:
            file: absolut file path to the file that should be loaded
        Returns:

        """
        fileName=file.name
        tables={}
        if fileName.endswith(self.FILE_TYPE):
            with ZipFile(file, mode="r") as documentZip:
                archivedFiles=documentZip.namelist()
                for archivedFile in archivedFiles:
                    with documentZip.open(archivedFile) as csvFile:
                        lod = CSV.fromCSV(csvFile.read().decode())
                        tableName=archivedFile[:-len(self.TABLE_TYPE)]
                        tables[tableName] = lod
        elif fileName.endswith(self.TABLE_TYPE):
            #single csv file load as sheet with one table
            lod = CSV.fromCSV(file.read().decode())
            tableName = fileName[:-len(self.TABLE_TYPE)]
            tables[tableName] = lod
        return tables


class ExcelDocument(SpreadSheet):
    """
    Provides methods to convert LoDs to an excel document and vice versa
    """

    FILE_TYPE = '.xlsx'
    MIME_TYPE = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'

    def __init__(self, name: str, engine:str=None,spreadSheetType:SpreadSheetType=SpreadSheetType.EXCEL):
        """
        Args:
            name(str): name of the document
        """
        super().__init__(name=name, spreadSheetType=spreadSheetType)
        if engine is None:
            engine="xlsxwriter"
        self.engine_kwargs={
                                'options': {
                                    'strings_to_numbers': True
                                }
                            }
        
        self.engine=engine
        
    def toBytesIO(self) -> BytesIO:
        """
        Converts the document into an BytesIO stream

        Returns:
            BytesIO Stream of the document
        """
        buffer = BytesIO()
       
        with pd.ExcelWriter(buffer,
                            engine=self.engine,
                            engine_kwargs=self.engine_kwargs,
                            ) as writer:
            for tableName, tableData in self.tables.items():
                df = pd.DataFrame(tableData)
                df.to_excel(writer,
                            sheet_name=tableName,
                            index=False)
        buffer.seek(0)
        buffer.name=self.filename
        return buffer

    def _loadFromBuffer(self,buffer):
        '''
        read my table from the given BytesIO buffer
        '''
        sheets = pd.read_excel(buffer, sheet_name=None).keys()
        tables={}
        for sheet in sheets:
            df = pd.read_excel(buffer, sheet_name=sheet, na_values=None)
            df=df.where(pd.notnull(df), None)
            lod=df.to_dict('records')
            # NaT handling issue due to a bug in pandas https://github.com/pandas-dev/pandas/issues/29024
            lod=[{k: v.to_pydatetime() if isinstance(v, Timestamp) else None if isinstance(v, type(NaT)) else v for k,v in d.items()} for d in lod]
            # float nan to None
            lod=[{k:v if not (isinstance(v, float) and math.isnan(v)) else None for k,v in d.items() }for d in lod]
            tables[sheet] = lod
        return tables

class OdsDocument(ExcelDocument):
    """
    OpenDocument Spreadsheet that can store multiple tables.
    Provides functions to traverse between LoD and ODS document
    """

    FILE_TYPE = '.ods'
    MIME_TYPE = 'application/vnd.oasis.opendocument.onlinespreadsheet'

    def __init__(self, name: str):
        """
        Args:
            name(str): name of the document
        """

        super().__init__(name=name, spreadSheetType=SpreadSheetType.ODS, engine="odf")
        self.engine_kwargs={}



'''
Created on 2022-04-30

@author: wf
'''
from spreadsheet.googlesheet import GoogleSheet
import pprint
from lodstorage.lod import LOD

class WikibaseQuery(object):
    '''
    a Query for Wikibase
    '''

    def __init__(self,entity:str,debug:bool=False):
        '''
        Constructor
        
        Args:
            entity(str): the entity this query represents
            debug(bool): if True switch on debugging
        '''
        self.debug=debug
        self.entity=entity
        self.propertiesByName={}
        self.propertiesById={}
        self.propertiesByVarname={}
        self.propertiesByColumn={}
        
    def addPropertyFromDescriptionRow(self,row):
        '''
        add a property from the given row
        
        Args:
            row(dict): the row to add
        '''
        propName=row['PropertyName']
        propId=row['PropertyId']
        column=row['Column']
        # properties might contain blank - replace for SPARQL variable names
        propVarname=propName.replace(" ","_")
        propVarname=propVarname.replace("-","_")
        row['PropVarname']=propVarname
        # set the values of the lookups
        self.propertiesByName[propName]=row
        self.propertiesByColumn[column]=row
        self.propertiesById[propId]=row
        self.propertiesByVarname[propVarname]=row
        
    def getColumnTypeAndVarname(self,propName):
        '''
        get a signature tuple consisting of columnName, propertType and SPARQL variable Name for the given property Name
        
        Args:
            propName(str): the name of the property
            
        Returns:
            column,propType,varName tupel
        '''
        column=self.propertiesByName[propName]["Column"]
        propType=self.propertiesByName[propName]["Type"]
        varName=self.propertiesByName[propName]["PropVarname"]
        return column,propType,varName
        
        
    def inFilter(self,values,propName:str="short_name",lang:str="en"):
        '''
        create a SPARQL IN filter clause
        
        Args:
            values(list): the list of values to filter for
            propName(str): the property name to filter with
            lang(str): the language to apply
        '''
        filterClause=f"\n  FILTER(?{propName} IN("
        delim=""
        for value in values:
            filterClause+=f"{delim}\n    '{value}'@{lang}"
            delim=","
        filterClause+="\n  ))."
        return filterClause
    
    def getValuesClause(self,values,propVarname:str="short_name",propType:str="text",lang:str=None,ignoreEmpty:bool=True,wbPrefix:str="http://www.wikidata.org/entity/"):
        '''
        create a SPARQL Values clause
        
        Args:
            values(list): the list of values to create values for
            propVarname(str): the property variable name to assign the values for
            ignoreEmpty(bool): ignore empty values if True
            wbPrefix(str): a wikibase/wikidata prefix to be removed for items values
        Returns:
            str: the SPARQL values clause
        '''
        valuesClause=f"\n  VALUES(?{propVarname}) {{"
        if lang is not None and propType=="text":
            lang=f'@{lang}'
        else:
            lang=''
        for value in values:
            if value or not ignoreEmpty:
                if not propType:
                    if value and value.startswith(wbPrefix):
                        value=value.replace(wbPrefix,"")
                    valuesClause+=f"\n   ( wd:{value} )"
                else:
                    # escape single quotes
                    value=value.replace("'","\\'")
                    valuesClause+=f"\n  ( '{value}'{lang} )"
        valuesClause+="\n  }."
        return valuesClause
        
    def asSparql(self,filterClause:str=None,orderClause:str=None,pk:str=None,lang:str="en"):
        '''
        get the sparqlQuery for this query optionally applying a filterClause
        
        Args:
            filterClause(str): a filter to be applied (if any)
            orderClause(str): an orderClause to be applied (if any)
            pk(str): primaryKey (if any)
            lang(str): the language to be used for labels
        '''
        sparql=f"""# 
# get {self.entity} records 
#  
PREFIX pq: <http://www.wikidata.org/prop/qualifier/>
PREFIX p: <http://www.wikidata.org/prop/>
PREFIX schema: <http://schema.org/>
PREFIX wd: <http://www.wikidata.org/entity/>
PREFIX wdt: <http://www.wikidata.org/prop/direct/>
PREFIX wikibase: <http://wikiba.se/ontology#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
SELECT ?item ?itemLabel ?itemDescription
"""
        for propVarname,row in self.propertiesByVarname.items():
            propValue=row["Value"]
            propType=row["Type"]
            # items will automatically fetch labels
            propLabel=f" ?{propVarname}Label" if not propType else ""
            # extid' will automatically fetch formatted URIs
            propUrl=f" ?{propVarname}Url" if propType=="extid" else ""
            if not propValue:
                sparql+=f"\n  ?{propVarname}{propLabel}{propUrl}"
        sparql+="""\nWHERE {
  ?item rdfs:label ?itemLabel.
  FILTER(LANG(?itemLabel) = "%s")
  OPTIONAL { 
    ?item schema:description ?itemDescription.
    FILTER(LANG(?itemDescription) = "%s")
  }
""" % (lang,lang)
        for propVarname,row in self.propertiesByVarname.items():
            propName=row["PropertyName"]
            propValue=row["Value"]
            propId=row["PropertyId"]
            propType=row["Type"]
            if propValue:
                sparql+=f"\n  ?item wdt:{propId} wd:{propValue}."
            else:
                # primary keys are not optional
                optional=pk is None or not propName==pk
                if optional:
                    sparql+=f"\n  OPTIONAL {{"
                sparql+=f"\n    ?item wdt:{propId} ?{propVarname}."
                if not propType:
                    sparql+=f"\n    ?{propVarname} rdfs:label ?{propVarname}Label."
                    sparql+=f"""\n    FILTER(LANG(?{propVarname}Label) = "{lang}")"""
                elif propType=="extid":
                    sparql+=f"\n    wd:{propId} wdt:P1630 ?{propVarname}FormatterUrl." 
                    sparql+=f"\n    BIND(IRI(REPLACE(?{propVarname}, '^(.+)$', ?{propVarname}FormatterUrl)) AS ?{propVarname}Url)."
                if optional:
                    sparql+=f"\n  }}"
        if filterClause is not None:
                sparql+=f"\n{filterClause}"        
        sparql+="\n}"
        if orderClause is not None:
            sparql+=f"\n{orderClause}"
        return sparql
            
    @classmethod
    def sparqlOfGoogleSheet(cls,url:str,sheetName:str,entityName:str,pkColumn:str,mappingSheetName="Wikidata",lang:str="en",debug:bool=False):
        '''
        get a sparql query for the given google sheet
        
        Args:
            url(str): the url of the sheet
            sheetName(str): the name of the sheet with the description
            entityName(str): the name of the entity as defined in the Wikidata mapping
            pkColumn(str): the column to use as a "primary key"
            mappingSheetName(str): the name of the sheet with the Wikidata mappings
            lang(str): the language to use (if any)
            debug(bool): if True switch on debugging
            
        Returns:
            WikibaseQuery
        '''
        queries=WikibaseQuery.ofGoogleSheet(url, mappingSheetName, debug)
        gs=GoogleSheet(url)
        gs.open([sheetName]) 
        lod=gs.asListOfDicts(sheetName)
        lodByPk,_dup=LOD.getLookup(lod,pkColumn)
        query=queries[entityName]
        propRow=query.propertiesByColumn[pkColumn]
        pk=propRow["PropertyName"]
        pkVarname=propRow["PropVarname"]
        pkType=propRow["Type"]
        valuesClause=query.getValuesClause(lodByPk.keys(),propVarname=pkVarname,propType=pkType,lang=lang)
        
        sparql=query.asSparql(filterClause=valuesClause,orderClause=f"ORDER BY ?{pkVarname}",pk=pk)
        return query,sparql
    
    @classmethod
    def ofGoogleSheet(cls,url:str,sheetName:str="Wikidata",debug:bool=False)->dict:
        '''
        create a dict of wikibaseQueries from the given google sheets row descriptions
        
        Args:
            url(str): the url of the sheet
            sheetName(str): the name of the sheet with the description
            debug(bool): if True switch on debugging
        '''
        gs=GoogleSheet(url)
        gs.open([sheetName])
        entityMapRows=gs.asListOfDicts(sheetName)
        return WikibaseQuery.ofMapRows(entityMapRows,debug=debug)
        
    @classmethod
    def ofMapRows(cls,entityMapRows:list,debug:bool=False):
        '''
        create a dict of wikibaseQueries from the given entityMap list of dicts
        
        Args:
            entityMapRows(list): a list of dict with row descriptions
            debug(bool): if True switch on debugging
        '''
        queries={}
        entityMapDict={}
        for row in entityMapRows:
            if "Entity" in row:
                entity=row["Entity"]
                if not entity in entityMapDict:
                    entityMapDict[entity]={}
                entityRows=entityMapDict[entity]
                if "PropertyName" in row:
                    propertyName=row["PropertyName"]
                    entityRows[propertyName]=row    
        if debug:
            pprint.pprint(entityMapDict)
        for entity in entityMapDict:
            wbQuery=WikibaseQuery.ofEntityMap(entity,entityMapDict[entity])
            queries[entity]=wbQuery
        return queries
    
    @classmethod
    def ofEntityMap(cls,entity:str,entityMap:dict):
        '''
        create a WikibaseQuery for the given entity and entityMap
        
        Args:
            entity(str): the entity name
            entityMap(dict): the entity property descriptions
        Returns:
            WikibaseQuery
        '''
        wbQuery=WikibaseQuery(entity)
        for row in entityMap.values():
            wbQuery.addPropertyFromDescriptionRow(row)
        return wbQuery


import pathlib
from setuptools import setup
from collections import OrderedDict
from spreadsheet.version import Version

here = pathlib.Path(__file__).parent.resolve()
readme= (here/"README.md").read_text()
requirements = (here / 'requirements.txt').read_text().split("\n")

setup(name=Version.name,
      version=Version.version,
      description=Version.description,
      long_description=readme,
      long_description_content_type='text/markdown',
      url=f'http://wiki.bitplan.com/index.php/{Version.name}',
      download_url=f'https://github.com/WolfgangFahl/{Version.name}',
      author='Wolfgang Fahl',
      author_email='wf@bitplan.com',
      license='Apache',
      project_urls=OrderedDict(
        (
            ("Code", f"https://github.com/WolfgangFahl/{Version.name}"),
            ("Issue tracker", f"https://github.com/WolfgangFahl/{Version.name}/issues"),
        )
      ),
      classifiers=[
            'Programming Language :: Python',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
      ],
      packages=['spreadsheet'],
      install_requires=requirements,
      zip_safe=False)



'''
Created on 2022-04-18

@author: wf
'''
from pathlib import Path
import json
import os
import re
import traceback
#from wikidataintegrator import wdi_core, wdi_login
from wikibaseintegrator import wbi_core, wbi_login, wbi_datatype
from lodstorage.lod import LOD
from lodstorage.sparql import SPARQL
import pprint
import dateutil.parser

class Wikidata:
    '''
    wikidata access
    
    see http://learningwikibase.com/data-import/
    '''
    
    def __init__(self,baseurl,debug:bool=False):
        '''
        Constructor
        
        Args:
            baseurl(str): the baseurl of the wikibase to use
            debug(bool): if True output debug information
        ''' 
        self.baseurl=baseurl
        self.debug=debug
        self.apiurl=f"{self.baseurl}/w/api.php"
        self.login=None
        self.user=None
        pass
    
    def getCredentials(self):
        '''
        get my credentials
        
        from the wd npm command line tool
        '''
        user=None
        pwd=None
        home = str(Path.home())
        configFilePath=f"{home}/.config/wikibase-cli/config.json"
        if os.path.isfile(configFilePath):
            with open(configFilePath, mode="r") as f:
                wikibaseConfigJson = json.load(f)
                credentials=wikibaseConfigJson["credentials"]
                if not self.baseurl in credentials:
                    raise Exception(f"no credentials available for {self.baseurl}")
                credentialRow=credentials[self.baseurl]
                user=credentialRow["username"]
                pwd=credentialRow["password"]
                pass
        return user,pwd
            
    def loginWithCredentials(self,user:str=None,pwd:str=None):
        '''
        login using the given credentials or credentials 
        retrieved via self.getCredentials
        
        Args:
            user(str): the username
            pwd(str): the password
        '''
        if user is None:
            user,pwd=self.getCredentials()       
            
        if user is not None:
            self.login = wbi_login.Login(user=user, pwd=pwd, mediawiki_api_url=self.apiurl)
            if self.login:
                self.user=user
            
    def logout(self):
        '''
        log the user out again
        '''
        self.user=None
        self.login=None
            
    def addItem(self,ist:list,label:str,description:str,lang:str="en",write:bool=True):
        '''
        Args:
            ist(list): item statements
            label(str): the english label
            description(str): the english description
            lang(str): the label language
            write(bool): if True do actually write
        '''
        wbPage=wbi_core.ItemEngine(data=ist,mediawiki_api_url=self.apiurl)
        wbPage.set_label(label, lang=lang)
        wbPage.set_description(description, lang=lang)
        if self.debug:
            pprint.pprint(wbPage.get_json_representation())
        if write and self.login:
            # return the identifier of the generated page
            return wbPage.write(self.login) # edit_summary=
        else:
            return None
            
    def getItemByName(self,itemName:str,itemType:str,lang:str="en"):
        '''
        get an item by Name
        
        Args:
            itemName(str): the item to look for
            itemType(str): the type of the item
            lang(str): the language of the itemName
        '''
        itemLabel=f'"{itemName}"@{lang}'
        sparqlQuery="""PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX wdt: <http://www.wikidata.org/prop/direct/>
    PREFIX wd: <http://www.wikidata.org/entity/>
    
    SELECT ?item ?itemLabel
    WHERE {
      {
        ?item wdt:P31 wd:%s.
        ?item rdfs:label ?itemLabel.
        # short name
        ?item wdt:P1813 %s
        FILTER(LANG(?itemLabel)= "%s" )
      } UNION {
        ?item wdt:P31 wd:%s.
        ?item rdfs:label ?itemLabel.
        FILTER(?itemLabel= %s )
      }
    }""" % (itemType,itemLabel,lang,itemType,itemLabel)
        endpointUrl="https://query.wikidata.org/sparql"
        sparql=SPARQL(endpointUrl)
        itemRows=sparql.queryAsListOfDicts(sparqlQuery)
        item=None
        if len(itemRows)>0:
            item=itemRows[0]["item"].replace("http://www.wikidata.org/entity/","")
        return item
            
    def addDict(self,row:dict,mapDict:dict,lang:str="en",write:bool=False,ignoreErrors:bool=False):
        '''
        add the given row mapping with the given map Dict
        
        Args:
            row(dict): the data row to add
            mapDict(dict): the mapping dictionary to use
            lang(str): the language for lookups
            write(bool): if True do actually write
            ignoreErrors(bool): if True ignore errors
        '''
        # dict of statements
        istMap={}
        # dict of errors
        errors={}
        propsByName,_dup=LOD.getLookup(mapDict.values(), "PropertyName", withDuplicates=False)
        for propId in mapDict.keys():
            propMap=mapDict[propId]
            column=propMap["Column"]
            colType=propMap["Type"]
            lookup=propMap["Lookup"]
            stToBeQualified=None
            if "Qualifier" in propMap:
                qualifierKey=propMap["Qualifier"]     
                if qualifierKey:
                    if not qualifierKey in propsByName: 
                        raise Exception(f"invalid qualifierKey {qualifierKey}")
                    else:
                        propIdToBeQualified=propsByName[qualifierKey]["PropertyId"]
                        if not propIdToBeQualified in istMap:
                            raise Exception(f"qualifierKey {qualifierKey} needs a statement to qualify")
                        stToBeQualified=istMap[propIdToBeQualified]         
            colValue=None
            try:
                if column:
                    if column in row:
                        colValue=row[column]
                else:
                    colValue=propMap["Value"]
                if colValue:
                    if lookup:
                        # ignore if the value is already a Wikibase Entity identifier of the form
                        # Q12345 ...
                        if not re.match(r"Q[0-9]+",colValue):
                            colValue=self.getItemByName(colValue, lookup, lang)
                if colValue and isinstance(colValue,str):
                    colValue=colValue.strip()
                st=None
                if colValue:
                    if colType=="year":
                        yearString=f"+{colValue}-01-01T00:00:00Z"
                        st=wbi_datatype.Time(yearString,prop_nr=propId,precision=9)
                    elif colType=="date":
                        dateValue = dateutil.parser.parse(colValue)
                        isoDate=dateValue.isoformat()
                        dateString=f"+{isoDate}Z"
                        st=wbi_datatype.Time(dateString,prop_nr=propId,precision=11)
                    elif colType=="extid":
                        st=wbi_datatype.ExternalID(value=colValue,prop_nr=propId)
                    elif colType=="string":
                        st=wbi_datatype.String(value=str(colValue),prop_nr=propId)
                    elif colType=="text":
                        st=wbi_datatype.MonolingualText(text=str(colValue),prop_nr=propId)
                    elif colType=="url":
                        st=wbi_datatype.Url(value=colValue,prop_nr=propId)
                    else:
                        st=wbi_datatype.ItemID(value=colValue,prop_nr=propId)
                    # qualifier for another statement?
                    if stToBeQualified is not None:
                        st.is_qualifier=True   
                        stToBeQualified.qualifiers.append(st)
                    else:    
                        istMap[propId]=st
            except Exception as ex:
                errors[column]=ex
                if self.debug:
                    print(traceback.format_exc())
        label=row["label"]
        description=row["description"]
        qid=None
        ist=list(istMap.values())
        if len(errors)==0 or ignoreErrors:
            qid=self.addItem(ist,label,description,write=write)
        return qid,errors
        
        
            
                